DEFAULT_DIAMETER = 30
DEFAULT_FLOW_THRESHOLD = 0.4
DEFAULT_CELLPROB_THRESHOLD = 0
DEFAULT_MIN_INTENSITY = 0
DEFAULT_MAX_INTENSITY = 2**8    # Max value in 8-bit image
DEFAULT_MIN_SIZE = 0
DEFAULT_MAX_SIZE = 2048*2048
DEFAULT_MIN_CIRCULARITY = 0
DEFAULT_MAX_CIRCULARITY = 1